private import util.vector;

// <<< memo >>>
// ubyte[] = [BB,GG,RR,AA]
//  > COLORREF
//  > OpenGL(GL_RGBA)
// ubyte[] = [RR,GG,BB,AA]
//  > D3DCOLOR
//  > DIB(RGBQUAD)
//  > OpenGL(GL_BGRA_EXT)
//  > Mac
//
// uint = 0xAABBGGRR
//  > COLORREF
//  > OpenGL(GL_RGBA) on windows
// uint = 0xRRGGBBAA
//  > OpenGL(GL_RGBA) on unix
// uint = 0xAARRGGBB
//  > D3DCOLOR
//  > DIB(RGBQUAD) on windows
//  > OpenGL(GL_BGRA_EXT) on windows
// uint = 0xBBGGRRAA
//  > DIB(RGBQUAD) on unix
//  > OpenGL(GL_BGRA_EXT) on unix
//  > Mac
//
// In this code, standard format for uint is 0xAABBGGRR.


// 0xAABBGGRR -> [RR,GG,BB,AA]
ubyte[] color2rgba(ubyte[] rgba, uint col) { rgba[0]=col&0xff; rgba[1]=(col>>8)&0xff; rgba[2]=(col>>16)&0xff; rgba[3]=(col>>24)&0xff; return rgba; }
// 0xAABBGGRR -> [BB,GG,RR,AA]
ubyte[] color2bgra(ubyte[] bgra, uint col) { bgra[0]=(col>>16)&0xff; bgra[1]=(col>>8)&0xff; bgra[2]=col&0xff; bgra[3]=(col>>24)&0xff; return bgra; }
// 0xAABBGGRR -> gray_scale
ubyte   color2gray(uint col) { return cast(ubyte)(((col&0xff)*77 + ((col>>8)&0xff)*150 + ((col>>16)&0xff)*29)>>8); }
// [RR,GG,BB,AA] -> 0xAABBGGRR
uint    rgba2color(ubyte[] rgba) { return rgba[0]|(rgba[1]<<8)|(rgba[2]<<16)|(rgba[3]<<24); }
// [BB,GG,RR,AA] -> 0xAABBGGRR
uint    bgra2color(ubyte[] bgra) { return bgra[2]|(bgra[1]<<8)|(bgra[0]<<16)|(bgra[3]<<24); }
// [RR,GG,BB] -> 0xffBBGGRR
uint    rgb2color(ubyte[] rgb) { return rgb[0]|(rgb[1]<<8)|(rgb[2]<<16)|0xff000000; }
// [BB,GG,RR] -> 0xffBBGGRR
uint    bgr2color(ubyte[] bgr) { return bgr[2]|(bgr[1]<<8)|(bgr[0]<<16)|0xff000000; }
// [RR,GG,BB] -> gray_scale
ubyte   rgb2gray(ubyte[] rgb) { return cast(ubyte)((rgb[0]*77 + rgb[1]*150 + rgb[2]*29)>>8); }
// [BB,GG,RR] -> gray_scale
ubyte   bgr2gray(ubyte[] bgr) { return cast(ubyte)((bgr[2]*77 + bgr[1]*150 + bgr[0]*29)>>8); }

uint    set_color(ubyte r, ubyte g, ubyte b, ubyte a)             { return r|(g<<8)|(b<<16)|(a<<24); }
ubyte[] set_rgb (ubyte[] dst, ubyte r, ubyte g, ubyte b)          { dst[0]=r; dst[1]=g; dst[2]=b;           return dst; }
ubyte[] set_rgba(ubyte[] dst, ubyte r, ubyte g, ubyte b, ubyte a) { dst[0]=r; dst[1]=g; dst[2]=b; dst[3]=a; return dst; }
ubyte[] set_bgr (ubyte[] dst, ubyte r, ubyte g, ubyte b)          { dst[0]=b; dst[1]=g; dst[2]=r;           return dst; }
ubyte[] set_bgra(ubyte[] dst, ubyte r, ubyte g, ubyte b, ubyte a) { dst[0]=b; dst[1]=g; dst[2]=r; dst[3]=a; return dst; }


version(LittleEndian)
{
ubyte[] copyLE3(ubyte[] dst, ubyte[] srcLE) { dst[]=srcLE[]; return dst; }
ubyte[] copyBE3(ubyte[] dst, ubyte[] srcBE) { dst[0]=srcBE[2]; dst[1]=srcBE[1]; dst[2]=srcBE[0]; return dst; }
ubyte[] copyLE4(ubyte[] dst, ubyte[] srcLE) { dst[]=srcLE[]; return dst; }
ubyte[] copyBE4(ubyte[] dst, ubyte[] srcBE) { dst[0]=srcBE[3]; dst[1]=srcBE[2]; dst[2]=srcBE[1]; dst[3]=srcBE[0]; return dst; }
}

version(BigEndian)
{
ubyte[] copyLE3(ubyte[] dst, ubyte[] srcLE) { dst[0]=srcLE[2]; dst[1]=srcLE[1]; dst[2]=srcLE[0]; return dst; }
ubyte[] copyBE3(ubyte[] dst, ubyte[] srcBE) { dst[]=srcBE[]; return dst; }
ubyte[] copyLE4(ubyte[] dst, ubyte[] srcLE) { dst[0]=srcLE[3]; dst[1]=srcLE[2]; dst[2]=srcLE[1]; dst[3]=srcLE[0]; return dst; }
ubyte[] copyBE4(ubyte[] dst, ubyte[] srcBE) { dst[]=srcBE[]; return dst; }
}




ubyte rgb2gray(int r, int g, int b) { return cast(ubyte)((r*77 + g*150 + b*29)>>8); }


void rgb2hsv(out int h, out int s, out int v, int r, int g, int b)
{
    long cmin,dc;

    if (r < g) {
        if (g < b) {
            cmin = r;
            v    = b;
            dc = v - cmin;
            if (dc == 0) {
                s = 0;
                h = 0;
            } else {
                s = (dc << 8) / v;
                h = (r-g)*60/dc + 240;
            }
        } else {
            cmin = (r < b) ? r : b;
            v    = g;
            dc = v - cmin;
            if (dc == 0) {
                s = 0;
                h = 0;
            } else {
                s = (dc << 8) / v;
                h = (b-r)*60/dc + 120;
            }
        }
    } else {
        if (r < b) {
            cmin = g;
            v    = b;
            dc = v - cmin;
            if (dc == 0) {
                s = 0;
                h = 0;
            } else {
                s = (dc << 8) / v;
                h = (r-g)*60/dc + 240;
            }
        } else {
            cmin = (g < b) ? g : b;
            v    = r;
            dc = v - cmin;
            if (dc == 0) {
                s = 0;
                h = 0;
            } else {
                s = (dc << 8) / v;
                h = (g-b)*60/dc;
                if (h<0) h+=360;
            }
        }
    }
}


void hsv2rgb(out int r, out int g, out int b, int h, int s, int v)
{
    long ch, cmin, hm;

    if (s == 0) {
        r = g = b = v;
    } else {
        cmin = (v*(255-s)) >> 8;
        ch = h/60;
        hm = (((v*s) >> 8) * (h - ch*60)) / 60;

        switch(ch){
        case 0:
            r = v;
            g = cmin + hm;
            b = cmin;
            break;
        case 1:
            r = v - hm;
            g = v;
            b = cmin;
            break;
        case 2:
            r = cmin;
            g = v;
            b = cmin + hm;
            break;
        case 3:
            r = cmin;
            g = v - hm;
            b = v;
            break;
        case 4:
            r = cmin + hm;
            g = cmin;
            b = v;
            break;
        case 5:
            r = v;
            g = cmin;
            b = v - hm;
            break;
        }
    }
}


void rgb2hls(out int h, out int l, out int s, int r, int g, int b)
{
    long cmax,cmin,dc,tc;

    if (r < g) {
        if (g < b) {
            cmin = r;
            cmax = b;
            dc = cmax - cmin;
            tc = cmax + cmin + 1;
            l = tc >> 1;
            if (dc == 0) {
                s = 0;
                h = 0;
            } else {
                if (l<128) s = (dc << 8) / tc;
                else s = (dc << 8) / (512-tc);
                h = (r-g)*60/dc + 240;
            }
        } else {
            cmin = (r < b) ? r : b;
            cmax = g;
            dc = cmax - cmin;
            tc = cmax + cmin + 1;
            l = tc >> 1;
            if (dc == 0) {
                s = 0;
                h = 0;
            } else {
                if (l<128) s = (dc << 8) / tc;
                else s = (dc << 8) / (512-tc);
                h = (b-r)*60/dc + 120;
            }
        }
    } else {
        if (r < b) {
            cmin = g;
            cmax = b;
            dc = cmax - cmin;
            tc = cmax + cmin + 1;
            l = tc >> 1;
            if (dc == 0) {
                s = 0;
                h = 0;
            } else {
                if (l<128) s = (dc << 8) / tc;
                else s = (dc << 8) / (512-tc);
                h = (r-g)*60/dc + 240;
            }
        } else {
            cmin = (g < b) ? g : b;
            cmax = r;
            dc = cmax - cmin;
            tc = cmax + cmin + 1;
            l = tc >> 1;
            if (dc == 0) {
                s = 0;
                h = 0;
            } else {
                if (l<128) s = (dc << 8) / tc;
                else s = (dc << 8) / (512-tc);
                h = (g-b)*60/dc;
                if (h<0) h+=360;
            }
        }
    }
}

void hls2rgb(out int r, out int g, out int b, int h, int l, int s)
{
    long  ch, ns, cmin, cmax, hm, dc;

    if (s == 0) {
        r = g = b = l;
    } else {
        if (l<128) dc = ((l<<1) * s) >> 8;
        else dc = ((512-(l<<1)) * s) >> 8;

        cmin = ((l<<1)-dc) >> 1;
        cmax = ((l<<1)+dc) >> 1;

        ch = h/60;
        hm = (dc * (h - ch*60)) / 60;
        
        switch(ch){
        case 0:
            r = cmax;
            g = cmin + hm;
            b = cmin;
            break;
        case 1:
            r = cmax - hm;
            g = cmax;
            b = cmin;
            break;
        case 2:
            r = cmin;
            g = cmax;
            b = cmin + hm;
            break;
        case 3:
            r = cmin;
            g = cmax - hm;
            b = cmax;
            break;
        case 4:
            r = cmin + hm;
            g = cmin;
            b = cmax;
            break;
        case 5:
            r = cmax;
            g = cmin;
            b = cmax - hm;
            break;
        }
    }
}




//--------------------------------------------------------------------------------
// simple color gradation class
//--------------------------------------------------------------------------------
class CGradation
{
private:
    vector4 acolGrad[256];


// constructor
//--------------------------------------------------
public:
    // default constructor
    this()
    {
        reset();
    }
    
    
    // copy constructor
    this(CGradation src)
    {
        copy(src);
    }
    
    
    // initialize constructor
    this(uint colStart, uint colEnd)
    {
        reset(colStart, colEnd);
    }
    
    


// operator
//--------------------------------------------------
public:
    // gradation[0-255] returns vector4*
    vector4* opIndex(int i)
    {
        return &acolGrad[i];
    }
    
    
    // gradation(0-1) returns vector4*
    vector4* opCall(float t)
    {
        return &acolGrad[(cast(int)(t*255.0f+0.5f)) & 255];
    }

    


// operation
//--------------------------------------------------
public:
    // reset
    CGradation reset()
    {
        set(&acolGrad[0],0,0,0,1);
        set(&acolGrad[255],1,1,1,1);
        for (int i=1; i<255; i++) { set(&acolGrad[i],0,0,0,-1); }
        return this;
    }
    

    // reset with start and end color
    CGradation reset(uint colStart, uint colEnd)
    {
        setUint(&acolGrad[0],  colStart);
        setUint(&acolGrad[255],colEnd);
        for (int i=1; i<255; i++) { set(&acolGrad[i],0,0,0,-1); }
        return this;
    }


    // copy
    CGradation copy(CGradation src)
    {
        for (int i=0; i<256; i++) {
            util.vector.copy(&acolGrad[i], &src.acolGrad[i]);
        }
        return this;
    }


    // add point
    CGradation add(float t, uint col)
    {
        int idx = (cast(int)(t*255.0f+0.5f)) & 255;
        setUint(&acolGrad[idx], col);
        return this;
    }


    // add point
    CGradation add(float t, float r, float g, float b, float a)
    {
        int idx = (cast(int)(t*255.0f+0.5f)) & 255;
        set(&acolGrad[idx], r, g, b, a);
        return this;
    }


    // add point
    CGradation add(float t, vector4* col)
    {
        int idx = (cast(int)(t*255.0f+0.5f)) & 255;
        util.vector.copy(&acolGrad[idx],col);
        return this;
    }


    // del point
    CGradation del(float t)
    {
        int idx = (cast(int)(t*255.0f+0.5f)) & 255;
        switch (idx) {
        case 0:
        case 255:
            set(&acolGrad[idx],0,0,0,1);
            break;
        default:
            set(&acolGrad[idx],0,0,0,-1);
            break;
        }
        return this;
    }
    

    // calculate gradation
    CGradation update()
    {
        int i, j;
        vector4* pcolStart, pcolEnd;
        float t, dt;
        for (i=0; i<256;) {
            for (j=i+1; j<256; j++) { if (acolGrad[j].a >= 0) break; }
            if (j==256) return this;
            pcolStart = &(acolGrad[i]);
            pcolEnd = &(acolGrad[j]);
            t = dt = 1.0f / cast(float)(j-i);
            for (i++; i<j; i++, t+=dt) { blend(&acolGrad[i], pcolStart, pcolEnd, t); }
        }
        return this;
    }


    // add blinking after update()
    CGradation addBlink(float start, float end, float interval, float r, float g, float b, float a)
    {
        int i, j;
        int istart = cast(int)(start*255.0f+0.5f);
        int iend   = cast(int)(end*255.0f+0.5f);
        int iinter = cast(int)(interval*255.0f+0.5f);
        if (iinter == 0) iinter = 1;
        
        for (i=istart; i<iend; i+=iinter*2) {
            for (j=0; j<iinter; j++) {
                set(&acolGrad[i+j], r, g, b, a);
            }
        }
        return this;
    }
};

