//------------------------------------------------------------------------------------------
// OpenGL Interleaved Array Class Family (Sprite, Charactor table)
//------------------------------------------------------------------------------------------
module util_opengl.array;


private import opengl;
private import openglu;
private import std.stdio;
private import std.math;
private import util_opengl.texture;
private import util.math;

private import vl = util.vector;




//----------------------------------------------------------------------------------------------------
//  glInterleavedArrays wrapping class
//----------------------------------------------------------------------------------------------------
class CArray
{
// interleaved array format
//--------------------------------------------------
protected:
    static enum data_format {
        SUPPORT_V = 0,
        SUPPORT_VT,
        SUPPORT_VN,
        SUPPORT_VNT,
        SUPPORT_VC,
        SUPPORT_VCT,
        max
    }

    static struct data_format_info {
        GLenum glformat;
        byte   size;
        byte   idxT;
        byte   idxC;
        byte   idxN;
        byte   idxV;
    }

    static data_format_info info[data_format.max] = [
        {glformat:GL_V3F,          size:4, idxT:0, idxC:0, idxN:0, idxV:0},
        {glformat:GL_T2F_V3F,      size:6, idxT:0, idxC:0, idxN:0, idxV:2},
        {glformat:GL_N3F_V3F,      size:7, idxT:0, idxC:0, idxN:0, idxV:3},
        {glformat:GL_T2F_N3F_V3F,  size:9, idxT:0, idxC:2, idxN:0, idxV:5},
        {glformat:GL_C4UB_V3F,     size:5, idxT:0, idxC:0, idxN:0, idxV:1},
        {glformat:GL_T2F_C4UB_V3F, size:7, idxT:0, idxC:0, idxN:2, idxV:3}
    ];


// member value
//--------------------------------------------------
protected:
    data_format format    = data_format.SUPPORT_V;  // data format
    float[]     data      = null;                   // data array
    int         data_size = 0;                      // data.length / info.size


// constructor
//--------------------------------------------------
protected:
    // default
    this() 
    {
    }


    // set format
    this(bool texcoord, bool normal, bool vcolor)
    {
        setFormat(texcoord, normal, vcolor);
    }


    // copy
    this(CArray arr)
    {
        copy(arr);
    }


// initialize
//--------------------------------------------------
public:
    // set format (if both normal and vcolor are ture, vcolor goes to false)
    CArray setFormat(bool texcoord, bool normal, bool vcolor)
    {
        vcolor &= !normal;
        format = cast(data_format)((texcoord?1:0) | (normal?2:0) | (vcolor?4:0));
        return this;
    }
    
    
    // allocate memory area
    CArray alloc(int size)
    {
        data.length = size * info[format].size;
        data_size = size;
        return this;
    }
    
    
    // expand memory area
    CArray expand(int size)
    {
        data.length = data.length + size * info[format].size;
        data_size += size;
        return this;
    }
    

    // copy
    CArray copy(CArray ary)
    {
        format = ary.format;
        data.length = ary.data.length;
        data_size = ary.data_size;
        data[] = ary.data[];
        return this;
    }


    // free
    void free()
    {
        data = null;
        data_size = 0;
    }


// reference
//--------------------------------------------------
    // get data
    float[] getData() { return data; }


    // get data count
    int getDataCount() { return data_size; }


    // support normal
    int isSupportNormal() { return ((cast(int)format) & 2); }


    // support vertex color
    int isSupportVertexColor() { return ((cast(int)format) & 4); }


    // suppert texture coordination
    int isSupportTexCoord() { return ((cast(int)format) & 1); }


    // get vertex
    vl.vector3* getVertex(int idx)
    {
        idx = idx * info[format].size + info[format].idxV;
        return cast(vl.vector3*)(data[idx..idx+3]);
    }


    // get normal
    vl.vector3* getNormal(int idx)
    {
        idx = idx * info[format].size + info[format].idxN;
        return cast(vl.vector3*)(data[idx..idx+3]);
    }


    // get vertex color
    ubyte* getColor(int idx)
    {
        idx = idx * info[format].size + info[format].idxC;
        return cast(ubyte*)(data[idx..idx+1]);
    }


    // get texture coordination
    vl.vector2* getTexCoord(int idx)
    {
        idx = idx * info[format].size + info[format].idxT;
        return cast(vl.vector2*)(data[idx..idx+2]);
    }


    // get next ... retrun null when finished
    vl.vector3* getNext(vl.vector3* v) { return cast(vl.vector3*)getNextPointer(&v.x); }
    vl.vector2* getNext(vl.vector2* v) { return cast(vl.vector2*)getNextPointer(&v.x); }
    ubyte*      getNext(ubyte* c)      { return cast(ubyte*)getNextPointer(cast(float*)c); }


    float* getNextPointer(float* ptr)
    {
        float* nextptr = &(ptr[info[format].size]);
        if (nextptr > &(data[$-1])) return null;
        return nextptr;
    }


// operation
//--------------------------------------------------
public:
    // multiply matrix to vertex
    CArray mul_vertex(vl.matrix4* m)
    {
        for (vl.vector3* v=getVertex(0); v!=null; v=getNext(v)) { vl.mul(v, m); }
        return this;
    }


    // multiply matrix to normal
    CArray mul_normal(vl.matrix4* m)
    {
        if (isSupportNormal()) {
            for (vl.vector3* v=getNormal(0); v!=null; v=getNext(v)) { vl.mul(v, m); }
        }
        return this;
    }
    

    // multiply quarternion
    CArray mul(vl.vector4* q)
    {
        vl.matrix4 m;
        mul_vertex(vl.loadRotate(&m, q));
        mul_normal(&m);
        return this;
    }


    // rotate z-axis to specifyed axis
    CArray setAxis(float x, float y, float z)
    {
        vl.vector4 qt;
        mul(vl.setQuart(&qt, -y, x, 0, acos(z)));
        return this;
    }


    // operation for vertex
    CArray operateVertex(void delegate(inout vl.vector3*) dg)
    {
        for (vl.vector3* v=getVertex(0); v!=null; v=getNext(v)) { dg(v); }
        return this;
    }


    // operation for normal
    CArray operateNormal(void delegate(inout vl.vector3*) dg)
    {
        if (isSupportNormal()) {
            for (vl.vector3* v=getNormal(0); v!=null; v=getNext(v)) { dg(v); }
        }
        return this;
    }


    // operation for vertex color
    CArray operateColor(void delegate(inout ubyte[]) dg)
    {
        if (isSupportVertexColor()) {
            for (ubyte* p=getColor(0); p!=null; p=getNext(p)) { 
                ubyte[] c=p[0..4];
                dg(c); 
            }
        }
        return this;
    }


    // operation for texture coordination
    CArray operateTexCoord(void delegate(inout vl.vector2*) dg)
    {
        if (isSupportTexCoord()) {
            for (vl.vector2* v=getTexCoord(0); v!=null; v=getNext(v)) { dg(v); }
        }
        return this;
    }


// drawing
//--------------------------------------------------
public:
    // glInterleavedArrays()
    CArray setInterleavedArrays()
    {
        glInterleavedArrays(info[format].glformat, info[format].size*float.sizeof, data.ptr);
        return this;
    }


    // glDrawArrays
    CArray draw(GLenum mode)
    {
        draw(mode, 0, data_size);
        return this;
    }


    // glDrawArrays
    CArray draw(GLenum mode, int idxStart, int count)
    {
        glDrawArrays(mode, idxStart, count);
        return this;
    }
}




//----------------------------------------------------------------------------------------------------
// Sprite class
//----------------------------------------------------------------------------------------------------
class CSprite : CArray
{
// member value
//--------------------------------------------------
public:
    CTexture       texture;
    vl.vector2*[4] uv;
    vl.vector3*[4] vertex;


// construct
//--------------------------------------------------
public:
    // set texture
    this(CTexture tex)
    {
        super(true, false, false);
        alloc(4);
        changeTexture(tex);
    }


    // copy
    this(CSprite arr)
    {
        copy(arr);
    }




// operation
//--------------------------------------------------
public:
    // allocate memory area
    CSprite alloc(int size)
    {
        format |= data_format.SUPPORT_VT;
        super.alloc(max(4,size));
        for (int i=0; i<4; i++) { uv[i] = getTexCoord(i); }
        for (int i=0; i<4; i++) { vertex[i] = getVertex(i); }
        return this;
    }
    

    // copy
    CSprite copy(CSprite spr)
    {
        super.copy(spr);
        changeTexture(spr.texture);
        return this;
    }


    // change texture
    CSprite changeTexture(CTexture tex)
    {
        texture = tex;
        if (!tex.isRegistered()) tex.register();
        return this;
    }


    // set UV coordination
    CSprite setUV(float u, float v, float w, float h)
    {
        vl.set(uv[0], u,   1.0f-v);
        vl.set(uv[1], u,   1.0f-v-h);
        vl.set(uv[2], u+w, 1.0f-v-h);
        vl.set(uv[3], u+w, 1.0f-v);
        return this;
    }


    // translate UV coordination
    CSprite translateUV(float u, float v)
    {
        vl.add(uv[0], u, -v);
        vl.add(uv[1], u, -v);
        vl.add(uv[2], u, -v);
        vl.add(uv[3], u, -v);
        return this;
    }


    // set vertex
    CSprite set(float x, float y, float w, float h, float z=0.0f)
    {
        vl.set(vertex[0], x,   y+h, z);
        vl.set(vertex[1], x,   y,   z);
        vl.set(vertex[2], x+w, y,   z);
        vl.set(vertex[3], x+w, y+h, z);
        return this;
    }


    // set vertex
    CSprite set(float w, float h)
    {
        float hw = w * 0.5f;
        float hh = h * 0.5f;
        vl.set(vertex[0], -hw,  hh, 0);
        vl.set(vertex[1], -hw, -hh, 0);
        vl.set(vertex[2],  hw, -hh, 0);
        vl.set(vertex[3],  hw,  hh, 0);
        return this;
    }


    // translation
    CSprite translate(float x, float y, float z=0.0f)
    {
        vl.add(vertex[0], x, y, z);
        vl.add(vertex[1], x, y, z);
        vl.add(vertex[2], x, y, z);
        vl.add(vertex[3], x, y, z);
        return this;
    }


    // scaling
    CSprite scale(float x, float y, float z=1.0f)
    {
        vl.scale(vertex[0], x, y, z);
        vl.scale(vertex[1], x, y, z);
        vl.scale(vertex[2], x, y, z);
        vl.scale(vertex[3], x, y, z);
        return this;
    }


    // rotate
    CSprite rotate(float r)
    {
        vl.vector2 rot;
        vl.setAngle(&rot, r);
        vl.mulZ(vertex[0], &rot);
        vl.mulZ(vertex[1], &rot);
        vl.mulZ(vertex[2], &rot);
        vl.mulZ(vertex[3], &rot);
        return this;
    }


    // multiply matrix
    CSprite mul(vl.matrix4* m)
    {
        vl.mul(vertex[0], m);
        vl.mul(vertex[1], m);
        vl.mul(vertex[2], m);
        vl.mul(vertex[3], m);
        return this;
    }
    

    // drawing
    CSprite draw()
    {
        setInterleavedArrays();
        texture.bind();
        super.draw(GL_TRIANGLE_FAN);
        texture.unbind();
        return this;
    }
}




//----------------------------------------------------------------------------------------------------
// Character table class
//----------------------------------------------------------------------------------------------------
class CCharTable
{
public:
    CSprite    sprite;          // sprite to display
    vl.vector2 sizeSpace;       // space size for each charactors on display
    vl.vector2 sizeCharactor;   // charactor size to print
    vl.vector2 sizeCharactorUV; // charactor size in UV
    int        widthTable;      // width of table
    int        offsetChar;      // index of head charactor


public:
    // constructor
    this(CTexture tex, int char_width, int char_height, int char_offset)
    {
        sprite = new CSprite(tex);
        changeTexture(tex, char_width, char_height, char_offset);
    }
    

public:
    // change texture
    CCharTable changeTexture(CTexture tex, int char_width, int char_height, int char_offset)
    {
        sprite.changeTexture(tex);
        vl.set(&sizeSpace, char_width, char_height);
        vl.copy(&sizeCharactor, &sizeSpace);
        vl.scaleinv(&sizeCharactorUV, &sizeCharactor, sprite.texture.width, sprite.texture.height);
        widthTable = cast(int)(sprite.texture.width/sizeCharactor.x + 0.5f);
        offsetChar = char_offset;
        return this;
    }
    

    // set spacing size
    CCharTable setSpace(float width, float height)
    {
        vl.set(&sizeSpace, width, height);
        return this;
    }
    

    // scaling charactor
    CCharTable scale(float x, float y)
    {
        vl.scale(&sizeCharactor, x, y);
        vl.scale(&sizeSpace,     x, y);
        return this;
    }
    

    // drawing
    CCharTable draw(float x, float y, char[] str)
    {
        int   idx, col, row;
        float u, v;
        float ix = x;
        y -= sizeCharactor.y;
        sprite.setInterleavedArrays();
        sprite.texture.bind();
        foreach (chr; str) {
            if (chr == '\n') { x=ix; y-=sizeSpace.y; }
            idx = chr - offsetChar;
            col = cast(int)(idx / widthTable);
            row = idx - col * widthTable;
            v = col * sizeCharactorUV.y;
            u = row * sizeCharactorUV.x;
            sprite.setUV(u, v, sizeCharactorUV.x, sizeCharactorUV.y);
            sprite.set  (x, y, sizeCharactor.x,   sizeCharactor.y);
            (cast(CArray)sprite).draw(GL_TRIANGLE_FAN);
            x += sizeSpace.x;
        }
        sprite.texture.unbind();
        return this;
    }
}




//----------------------------------------------------------------------------------------------------
// Circle array class
//----------------------------------------------------------------------------------------------------
class CCircleArray : CArray
{
// member value
//--------------------------------------------------
public:
    vl.vector3*[]  vert_inner;
    vl.vector3*[]  vert_outer;
    vl.vector2[]   rot_table;
    int            div_count;


// construct
//--------------------------------------------------
public:
    // constructor
    this(int div_count_, bool texcoord=true, bool normal=false, bool vcolor=false) 
    {
        super(texcoord, normal, vcolor);
        alloc(div_count_);
    }
    
    // copy constructor
    this(CCircleArray arr) { copy(arr); }


// operation
//--------------------------------------------------
public:
    // allocate by fan dividing count (>=3)
    CCircleArray alloc(int div_count_)
    {
        div_count = max(3,div_count_);
        super.alloc((div_count+1)*2);
        vert_inner = new vl.vector3*[div_count+1];
        vert_outer = new vl.vector3*[div_count+1];
        rot_table  = new vl.vector2[div_count+1];
        double r  = 0;
        double dr = dRD/div_count;
        for (int i=0; i<div_count+1; i++) {
            vert_inner[i] = getVertex(i*2);
            vert_outer[i] = getVertex(i*2+1);
            vl.setAngle(&rot_table[i], r);
            r+=dr;
        }
        return this;
    }
    

    // copy
    CCircleArray copy(CCircleArray arr)
    {
        super.copy(arr);
        div_count = arr.div_count;
        return this;
    }


    // set UV coordination
    CCircleArray setUV(float u, float v, float w, float h)
    {
        if (!isSupportTexCoord()) return this;
        float du=w/div_count;
        v = 1-v;
        vl.vector2* uv = getTexCoord(0);
        for (int i=0; i<div_count+1; i++) {
            uv=getNext(vl.set(uv, u, v));
            uv=getNext(vl.set(uv, u, v-h));
            u+=du;
        }
        return this;
    }


    // set vertex color
    CCircleArray setColor(uint col_inner, uint col_outer)
    {
        if (!isSupportVertexColor()) return this;
        ubyte* col = getColor(0);
        ubyte[4] cin;  cin[]  = getColor(col_inner)[0..4];
        ubyte[4] cout; cout[] = getColor(col_outer)[0..4];
        for (int i=0; i<div_count+1; i++) {
            col[0..4] = cin[];
            col=getNext(col);
            col[0..4] = cout[];
            col=getNext(col);
        }
        return this;
    }


    // set raduis
    CCircleArray setRaduis(float in_r, float out_r, float in_z=0, float out_z=0)
    {
        for (int i=0; i<div_count+1; i++) {
            vl.scale(&vert_inner[i].v2, &rot_table[i], in_r);
            vl.scale(&vert_outer[i].v2, &rot_table[i], out_r);
            vert_inner[i].z=in_z;
            vert_outer[i].z=out_z;
        }
        return this;
    }


    // set cylinderical
    CCircleArray setCylinder(float r, float min_z=0, float max_z=0)
    {
        return setRaduis(r, r, max_z, min_z);
    }
    

    // set inner raduis
    CCircleArray setInnerRaduis(float in_r, float in_z=0)
    {
        for (int i=0; i<div_count+1; i++) {
            vl.scale(&vert_inner[i].v2, &rot_table[i], in_r);
            vert_inner[i].z=in_z;
        }
        return this;
    }


    // set outer raduis
    CCircleArray setOuterRaduis(float out_r, float out_z=0)
    {
        for (int i=0; i<div_count+1; i++) {
            vl.scale(&vert_outer[i].v2, &rot_table[i], out_r);
            vert_outer[i].z=out_z;
        }
        return this;
    }


    // drawing
    CCircleArray draw()
    {
        setInterleavedArrays();
        super.draw(GL_TRIANGLE_STRIP);
        return this;
    }
}




//----------------------------------------------------------------------------------------------------
// Strip array class
//----------------------------------------------------------------------------------------------------
class CStripArray : CArray
{
// member value
//--------------------------------------------------
public:
    vl.vector3*[]  vert0;
    vl.vector3*[]  vert1;
    int         joint_count;
    int         idxStart;
    int         idxCount;


// construct
//--------------------------------------------------
public:
    // constructor
    this(int joint_count_) { alloc(joint_count_); }
    this(CStripArray arr)  { copy(arr); }


// operation
//--------------------------------------------------
public:
    // allocate by joint count
    CStripArray alloc(int joint_count_)
    {
        joint_count = max(1,joint_count_);
        super.alloc((joint_count+2)*2);
        vert0 = new vl.vector3*[joint_count+2];
        vert1 = new vl.vector3*[joint_count+2];
        for (int i=0; i<joint_count+2; i++) {
            vert0[i] = getVertex(i*2);
            vert1[i] = getVertex(i*2+1);
        }
        return this;
    }
    
    // copy
    CStripArray copy(CStripArray arr)
    {
        super.copy(arr);
        joint_count = arr.joint_count;
        idxStart = arr.idxStart;
        idxCount = arr.idxCount;
        return this;
    }

    // set UV coordination
    CStripArray setUV(float u, float v, float w, float h, float t0, float t1)
    {
        if (!isSupportTexCoord()) return this;

        idxStart = (t0==0)?2:0;
        idxCount = data_size - idxStart - ((t1==1)?2:0);
        
        float du=w*(t1-t0)/joint_count;
        float u_end=u+w;
        v = 1-v;
        vl.vector2* uv = getTexCoord(0);
        
        // start point
        uv=getNext(vl.set(uv, u, v));
        uv=getNext(vl.set(uv, u, v-h));
        u+=w*t0;
        
        // body
        for (int i=1; i<=joint_count; i++) {
            uv=getNext(vl.set(uv, u, v));
            uv=getNext(vl.set(uv, u, v-h));
            u+=du;
        }
        
        // end point
        uv=getNext(vl.set(uv, u_end, v));
        uv=getNext(vl.set(uv, u_end, v-h));
        
        return this;
    }

    // set strip
    CStripArray setStrip(vl.vector2[] vcJoint, float width)
    {
        int imax = min(joint_count, cast(int)vcJoint.length);
        assert(imax < 2);
        
        vl.vector2[2] v;
        vl.vector2    dir;
        int           sw = 0;

        vl.calcDirection(&v[0], &vcJoint[0], &vcJoint[1]);

        // start point
        vl.add(&vert0[0].v2, vl.add(vl.scale(vl.set(&vert0[1].v2,  v[0].y, -v[0].x), width), &vcJoint[0]), &v[0]);
        vl.add(&vert1[0].v2, vl.add(vl.scale(vl.set(&vert1[1].v2, -v[0].y,  v[0].x), width), &vcJoint[0]), &v[0]);
        
        // body
        for (int i=1; i<imax-1; i++) {
            sw=1-sw;
            vl.scale(vl.add(&dir, vl.calcDirection(&v[sw], &vcJoint[i], &vcJoint[i+1]), &v[1-sw]), 0.5);
            vl.add(vl.scale(vl.set(&vert0[i+1].v2,  dir.y, -dir.x), width), &vcJoint[i]);
            vl.add(vl.scale(vl.set(&vert1[i+1].v2, -dir.y,  dir.x), width), &vcJoint[i]);
        }
        
        // end point
        vl.sub(&vert0[imax+1].v2, vl.add(vl.scale(vl.set(&vert0[imax].v2,  v[sw].y, -v[sw].x), width), &vcJoint[imax-1]), &v[sw]);
        vl.sub(&vert1[imax+1].v2, vl.add(vl.scale(vl.set(&vert1[imax].v2, -v[sw].y,  v[sw].x), width), &vcJoint[imax-1]), &v[sw]);
        
        return this;
    }

    // set strip
    CStripArray setStrip(vl.vector2[] vcJoint, vl.vector2[] vcDir, float width)
    {
        int imax = min(joint_count, cast(int)vcJoint.length);
        assert(imax < 2);
        
        // start point
        vl.add(&vert0[0].v2, vl.add(vl.scale(vl.set(&vert0[1].v2,  vcDir[0].y, -vcDir[0].x), width), &vcJoint[0]), &vcDir[0]);
        vl.add(&vert1[0].v2, vl.add(vl.scale(vl.set(&vert1[1].v2, -vcDir[0].y,  vcDir[0].x), width), &vcJoint[0]), &vcDir[0]);
        
        // body
        for (int i=1; i<imax-1; i++) {
            vl.add(vl.scale(vl.set(&vert0[i+1].v2,  vcDir[i].y, -vcDir[i].x), width), &vcJoint[i]);
            vl.add(vl.scale(vl.set(&vert1[i+1].v2, -vcDir[i].y,  vcDir[i].x), width), &vcJoint[i]);
        }
        
        // end point
        vl.sub(&vert0[imax+1].v2, vl.add(vl.scale(vl.set(&vert0[imax].v2,  vcDir[imax-1].y, -vcDir[imax-1].x), width), &vcJoint[imax-1]), &vcDir[imax-1]);
        vl.sub(&vert1[imax+1].v2, vl.add(vl.scale(vl.set(&vert1[imax].v2, -vcDir[imax-1].y,  vcDir[imax-1].x), width), &vcJoint[imax-1]), &vcDir[imax-1]);
        
        return this;
    }

    // drawing
    CStripArray draw()
    {
        setInterleavedArrays();
        super.draw(GL_TRIANGLE_STRIP, idxStart, idxCount);
        return this;
    }
}





