//--------------------------------------------------------------------------------
//  SDL OpenGL Main Fiber class for D v0.1.0
//
//    Copyright  (C)  2006-2007  kei mesuda
//    (http://www.yomogi.sakura.ne.jp/~si)
//--------------------------------------------------------------------------------
module util_sdl.mainfiber;


private import std.string;

private import sdl;
version(use_mixer)  { private import sdl_mixer; }

private import util.list;
private import util_sdl.fiber;
private import util_sdl.controler;
version(use_opengl) { 
    private import opengl;
    private import util_opengl.expansion;
}






//--------------------------------------------------------------------------------
// Global values
//--------------------------------------------------------------------------------
public:
FPSModule       moduleFPS;      // FPS manager module
FiberModule     moduleFiber;    // fiber manager module
ControlModule   moduleControl;  // controler manager module

private:
SDL_SETTING     settingSDL;     // setting
SDL_Surface*    surfaceScreen;  // pointer of screen surface






//--------------------------------------------------------------------------------
// SDL setting infomation structure
//--------------------------------------------------------------------------------
public struct SDL_SETTING
{
    // OpenGL settings
    struct OPENGL
    {
        bool    isReferSetting = false;     // Dose OpenGL refer this setting ?
        
        int     countRBit = 8;              // Bit count for Red   channel
        int     countGBit = 8;              // Bit count for Green channel
        int     countBBit = 8;              // Bit count for Blue  channel
        int     countABit = 8;              // Bit count for Alpha channel

        int     sizeDepthBuffer = 32;       // Size of Z buffer
        int     sizeStencilBuffer = 0;      // Size of stencil buffer
        bool    isDoubleBuffered = true;    // Is screen double bufferd ?

        int     countRAccum = 0;            // Bit count for Red   accumlation buffer
        int     countGAccum = 0;            // Bit count for Green accumlation buffer
        int     countBAccum = 0;            // Bit count for Blue  accumlation buffer
        int     countAAccum = 0;            // Bit count for Alpha accumlation buffer
    }

    // Window settings
    char[]          title            = "";       // Window title
    bool            isFullScreen     = false;    // Is screen full ?
    bool            isCursorVisible  = true;     // Is cursor visible ?

    // Video settings
    int             width            = 640;      // Screen width
    int             height           = 480;      // Screen height
    int             countBitPerPixel = 32;       // Bit count for 1 pixel
    ulong           flagInitialize   = SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK;     // flags for initializing

    // Sound settings
    SDL_AudioSpec Audio = {
        freq: 22050,
        format: AUDIO_S16,
        channels: 2,
        samples: 1024,
        callback: null, // set in initializeSDL()
        userdata: null
    };
    void delegate(short[]) dgSoundCallBack = null;
    
    // OpenGL setting
    OPENGL          OpenGL;
}






//--------------------------------------------------------------------------------
//  Main fiber class
//--------------------------------------------------------------------------------
public abstract class IMainFiber : public IFiber
{
//----------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------
public:
    this()
    {
        // Initialize global modules
        moduleFPS     = new FPSModule;
        moduleFiber   = new FiberModule;
        moduleControl = new ControlModule;
        
        // Add main fiber to the fiber list
        moduleFiber.add(this);
    }
    



//----------------------------------------------------------------------
// Execution
//----------------------------------------------------------------------
public:
    void execute(char[][] arg)
    {
        // initialize
        enterInstance(arg);                 // Call enterInstance()
        initializeSDL();                    // Call initializeSDL()
        moduleControl.updateJoystickList(); // Update joy stick list
        moduleFPS.reset();                  // Reset ticks
        moduleFiber.enter_all();            // Call all IFiber.enter()
        moduleFiber.enter_scene_all();      // Call all IFiber.enterScene()

        
        // execution
        while (true) {
            // Event loop
            SDL_Event   msgSDLEvent;
            while (SDL_PollEvent(&(msgSDLEvent)) && !moduleFiber.isHalted()) {
                // update keyboard table
                moduleControl.update_keytable();
                
                // call IMainFiber.eventHandler()
                eventHandler(&(msgSDLEvent));

                // system halt
                if (moduleControl.is_system_halted() || msgSDLEvent.type == SDL_QUIT) {
                    moduleFiber.halt();
                }
            }

            // Exit loop when halt flag is standing
            if (moduleFiber.isHalted()) break;
            
            // Update control key counter
            moduleControl.update_counter();

            // Synchronize FPS
            moduleFPS.sync();

            // Call all IFiber.run()
            moduleFiber.run_all();

            // Call all IFiber.draw() if drawable
            if (!moduleFPS.isSkipping()) {
                moduleFiber.draw_all();

                // Swap Buffers
                version(use_opengl) SDL_GL_SwapBuffers();
                else                SDL_Flip(IMainFiber.getScreenSurface());
            }
        }

        
        // finalize
        scope(exit) {
            moduleFiber.exit_scene_all();   // Call all IFiber.exitScene()
            moduleFiber.exit_all();         // IFiber.exit()
            terminateSDL();                 // Call terminateSDL()
            exitInstance();                 // Call exitInstance()
        }
    }




//----------------------------------------------------------------------
//  Virtual functions; Implement these functions.
//    These functions are called from main loop properly.
//----------------------------------------------------------------------
public:
    // Main loop calls this first of all with arguments.
    void enterInstance(char[][] arg)
    {
        enterInstance();
    }


    // Main loop calls this first of all.
    void enterInstance()
    {
    }
    
    
    // Main loop calls this at the end.
    void exitInstance()
    {
    }
    
    
    // Main loop calls this when SDL_Event is in the event que. When this returns false, main loop halts.
    void eventHandler(SDL_Event* pEvent)
    {
    }
    
    
    // And also virtual functions of IFiber
    // void enter()
    // void exit()
    // void enterScene()
    // void exitScene()
    // void run()
    // void draw()
    

    // default sound call back
    void callbackSound(short[] stream)
    {
    }




//  Access to setting information
//----------------------------------------------------------------------
public:
    // Window setting
    static void setWindow(char[] title, int width, int height, int fps=60, int maxskip=5)
    {
        settingSDL.title  = title;
        settingSDL.width  = width;
        settingSDL.height = height;
        moduleFPS.setFPS(fps, maxskip);
    }
    
    
    // Set sound call back
    static void setSoundCallBack(void delegate(short[]) dg)
    {
        settingSDL.dgSoundCallBack = dg;
    }
    
    
    // Use accumlation buffer
    static void useAccumBuffer()
    {
        settingSDL.OpenGL.countRAccum = settingSDL.OpenGL.countRBit;
        settingSDL.OpenGL.countGAccum = settingSDL.OpenGL.countGBit;
        settingSDL.OpenGL.countBAccum = settingSDL.OpenGL.countBBit;
        settingSDL.OpenGL.countAAccum = settingSDL.OpenGL.countABit;
    }
    
    
    // get screen width
    static int getScreenWidth()
    {
        return settingSDL.width;
    }


    // get screen height
    static int getScreenHeight()
    {
        return settingSDL.height;
    }


    // Reference for frame buffer size [byte]
    static int getFrameBufferSize()    
    {
        return ((settingSDL.OpenGL.countRBit + 
                 settingSDL.OpenGL.countGBit + 
                 settingSDL.OpenGL.countBBit + 
                 settingSDL.OpenGL.countABit + 7) & 0xf8);
    }


    // Parse arguments (-full / -res w h)
    static void parseArg(char[][] arg, void delegate(inout int, char[][]) dg=null)
    {
        for (int i=1; i<arg.length; ++i) {
            switch (arg[i]) {
            case "-full":
                settingSDL.isFullScreen = true;
                break;
            case "-res":
                if (i<arg.length-2) {
                    settingSDL.width  = atoi(arg[++i]);
                    settingSDL.height = atoi(arg[++i]);
                }
                break;
            default:
                if (dg != null) dg(i, arg);
                break;
            }
        }
    }




//  drawing
//----------------------------------------------------------------------
public:
    // Get pointer of screen surface
    static SDL_Surface* getScreenSurface()
    {
        return surfaceScreen;
    }

    
    // Create compatible SDL_Surface
    static SDL_Surface* createSurface(int width, int height, bool useSystemMemory = false)
    {
        SDL_PixelFormat* form = getScreenSurface().format;
        return SDL_CreateRGBSurface((useSystemMemory) ? (SDL_SWSURFACE) : (SDL_HWSURFACE), width, height, form.BitsPerPixel,
                                    form.Rmask, form.Gmask, form.Bmask, form.Amask);
    }


    // Capture screen
    static void captureScreen(char[] nameFile)
    {
        SDL_SaveBMP(surfaceScreen, toStringz(nameFile));
    }
    
    
    // Open GL function
    version(use_opengl) {
        // Double buffering ?
        static bool isUsingDepthBuffer()
        {
            return (settingSDL.OpenGL.sizeDepthBuffer > 0);
        }


        // Stencil buffering ?
        static bool isUsingStencilBuffer()
        {
            return (settingSDL.OpenGL.sizeStencilBuffer > 0);
        }


        // Alpha channel ?
        static bool isUsingAlphaChannel()
        {
            return (settingSDL.OpenGL.countABit > 0);
        }
    }
    



//  audio
//----------------------------------------------------------------------
public:
    // load wave file and convert format
    static void loadWave(char[] nameFile)
    {
        SDL_AudioSpec spc;
        SDL_AudioCVT  cvt;
        ubyte* dat;
        uint   len;
        
        // load file
        if (SDL_LoadWAV(nameFile.ptr, &spc, &dat, &len) == null) {
            throw new Exception(nameFile ~ "; cannot read");
        }
        
        // convert
        SDL_AudioSpec* dst = &settingSDL.Audio;
        if (SDL_BuildAudioCVT(&cvt, spc.format, spc.channels, spc.freq, dst.format, dst.channels, dst.freq) < 0) {
            throw new Exception(nameFile ~ "; cannot convert");
        }
        cvt.buf = (new ubyte[cvt.len*cvt.len_mult]).ptr;
        cvt.buf[0..len] = dat[0..len];
        SDL_ConvertAudio(&cvt);
    }
    
    
    


// Functions call from main routines
//--------------------------------------------------
// Initialize and terminate SDL; Possible to override but it's not necessary.
//----------------------------------------------------------------------
protected:
    void initializeSDL()
    {
        // Initialize SDL
        if (SDL_Init(settingSDL.flagInitialize) < 0) {
            throw new Exception("Failed to initialize SDL.");
        }


        // Initialize SDL_mixer
        version(use_mixer) {
            // open SDL Mixer
            if (Mix_OpenAudio(settingSDL.Audio.freq, 
                              settingSDL.Audio.format, 
                              settingSDL.Audio.channels,
                              settingSDL.Audio.samples) < 0) {
                throw new Exception("Failed to initialize SDL Mixer.");
            }
        } else {
            // set default call back function
            if (settingSDL.Audio.callback == null) {
                settingSDL.Audio.callback = &defaultAudioCallBack;
            }
            if (settingSDL.dgSoundCallBack == null) {
                settingSDL.dgSoundCallBack = &callbackSound;
            }
            // open SDL Audio
            if (SDL_OpenAudio(&settingSDL.Audio, null) < 0) {
                throw new Exception("Failed to initialize SDL Audio.");
            }
        }


        // OpenGL setting
        version(use_opengl) {
            if (settingSDL.OpenGL.isReferSetting) {
                SDL_GL_SetAttribute(SDL_GL_RED_SIZE,     settingSDL.OpenGL.countRBit);
                SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,   settingSDL.OpenGL.countGBit);
                SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,    settingSDL.OpenGL.countBBit);
                SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,   settingSDL.OpenGL.countABit);
                SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE,  getFrameBufferSize());

                SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, cast(int)settingSDL.OpenGL.isDoubleBuffered);
                SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,   settingSDL.OpenGL.sizeDepthBuffer);
                SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, settingSDL.OpenGL.sizeStencilBuffer);

                SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE,   settingSDL.OpenGL.countRAccum);
                SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE, settingSDL.OpenGL.countGAccum);
                SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE,  settingSDL.OpenGL.countBAccum);
                SDL_GL_SetAttribute(SDL_GL_ACCUM_ALPHA_SIZE, settingSDL.OpenGL.countAAccum);
            }

            // Video mode flags with OpenGL
            ulong   flagVideoMode = SDL_HWSURFACE | SDL_OPENGL;
        } else {
        
            // Video mode flags without OpenGL
            ulong   flagVideoMode = SDL_HWSURFACE | SDL_DOUBLEBUF;
        }


        // Initialize video
        if (settingSDL.isFullScreen) flagVideoMode |= SDL_FULLSCREEN;
        surfaceScreen = SDL_SetVideoMode(settingSDL.width, settingSDL.height, settingSDL.countBitPerPixel, flagVideoMode);
        if (!surfaceScreen) throw new Exception("Failed to initialize video.");


        // Others
        SDL_ShowCursor((settingSDL.isCursorVisible) ? SDL_ENABLE : SDL_DISABLE);    // Cursor
        SDL_WM_SetCaption(settingSDL.title.ptr, settingSDL.title.ptr);              // Window caption
        

        // Initialize OpenGL
        version(use_opengl) {
            if (isUsingDepthBuffer())   glEnable (GL_DEPTH_TEST);
            else                        glDisable(GL_DEPTH_TEST);
            if (isUsingStencilBuffer()) glEnable (GL_STENCIL_TEST);
            else                        glDisable(GL_STENCIL_TEST);
            if (isUsingAlphaChannel())  glEnable (GL_BLEND);
            else                        glDisable(GL_BLEND);

            glViewport(0, 0, settingSDL.width, settingSDL.height);
            setScreenPers(45, cast(float)settingSDL.width, cast(float)settingSDL.height, -100, 1000);
            glLoadIdentity();
        }
    }


    void terminateSDL()
    {
        // Show cursor
        SDL_ShowCursor(SDL_ENABLE);

        // Close SDL_mixer / SDL audio
        version(use_mixer) Mix_CloseAudio();
        else SDL_CloseAudio();

        // Free screen surface
        SDL_FreeSurface(surfaceScreen);

        // Shut down SDL
        SDL_Quit();
    }
}




// default SDL Audio sound call back
extern (C) {
void defaultAudioCallBack(void* usrdata, ubyte* stream, int len) {
    settingSDL.dgSoundCallBack((cast(short*)stream)[0..len>>1]);
}
}






//--------------------------------------------------------------------------------
// FPS Managing module
//--------------------------------------------------------------------------------
public class FPSModule
{
//  Values
//--------------------------------------------------
private:
    float    timeRefleshInterval   = 16.666666f; // Refleshing interval [ms]
    int      countMaxSkippingFrame = 0;          // Max skipping frame count
    int      countFPSAveragingFrame= 30;         // Averaging frame count
    float    fpsAverage            = 60.0f;      // Average reflesh rate
    bool     flagSkipping          = false;      // frame skipping ?
    bool     flagSlowing           = false;      // slow ?

    int      counterSkippingFrame     = 0;       // if counterSkippingFrame != 0, skip draw()
    long     counterFPSAveragingFrame = 0;       // Averaging frame counter
    float    timeOdd               = 0.0f;       // Odd of synchronizing wait time
    long     tickPrevAvaraging     = 0;          // Previous averaging time
    long     tickPrevSync          = 0;          // Previous synchronizing time
    long     tickPrevMeasure       = 0;          // Previous measure time


//  Constructor
//--------------------------------------------------
public:
    this()
    {
    }


//  Functions
//--------------------------------------------------
public:
    // set FPS and limit of skipping frame. When fps sets to 0, execute as fast as possible.
    void setFPS(int fps, int maxskip = 0)
    {
        fpsAverage = cast(float)fps;

        if (fps == 0) {
            // Execute as fast as possible
            timeRefleshInterval   = 0;
            countMaxSkippingFrame = 0;
        } else {
            timeRefleshInterval   = 1000.0f / cast(float)fps;
            countMaxSkippingFrame = (maxskip < 0) ? (0) : (maxskip);
        }
    }


    // set FPS averaging frame
    void setFPSAveragingFrame(int frame)
    {
        countFPSAveragingFrame = (frame < 1) ? 1 : frame;
    }


    // get FPS
    float getFPS()
    {
        return fpsAverage;
    }
    

    // skipping ?
    bool isSkipping()
    {
        return flagSkipping;
    }


    // slowing ?
    bool isSlowing()
    {
        return flagSlowing;
    }


    // measure tick interval from previous call [ms]
    long measure()
    {
        long tickCurrent  = SDL_GetTicks();
        long tickInterval = tickCurrent - tickPrevMeasure;
        tickPrevMeasure = tickCurrent;
        return tickInterval;
    }




//  Functions call from main routines
//--------------------------------------------------
private:
    // reset ticks
    void reset()
    {
        tickPrevAvaraging = SDL_GetTicks();
        tickPrevSync      = SDL_GetTicks();
        tickPrevMeasure   = SDL_GetTicks();
    }


    // synchronize
    void sync()
    {
        // Current tick count
        long tickCurrent = SDL_GetTicks();
        
        // Calclate waiting time
        timeOdd += timeRefleshInterval - cast(float)(tickCurrent - tickPrevSync);

        // Skipping ?
        if (timeOdd > 0) {
            // Wait (no skipping)
            flagSkipping = false;                   // No skipping
            flagSlowing = false;                    // No slowing
            counterSkippingFrame = 0;               // Do not skip draw()
            int tickOdd = cast(int)timeOdd;         // calculate waiting time
            if (tickOdd != 0) SDL_Delay(tickOdd);   // Waiting
            _calc_fps_average();                    // Calculate average of FPS

        } else {
            // No wait (skipping)
            flagSkipping = true;                    // skipping
            ++counterSkippingFrame;                 // Increase skipping frame counter
            
            // Cannot skip draw() anymore
            if (counterSkippingFrame > countMaxSkippingFrame) {
                flagSlowing = true;                 // slowing
                counterSkippingFrame = 0;
                timeOdd = 0.0f;                     // Reset waiting time
                _calc_fps_average();                // Calculate average of FPS
            }
            
            // When draw() is skipped, there are nothing to do.
        }
        
        // Next time's tickPrevious = this time's tickCurrent
        tickPrevSync = tickCurrent;
    }


    // calculate average FPS
    void _calc_fps_average()
    {
        if (++counterFPSAveragingFrame == countFPSAveragingFrame) {
            long tickCurrent  = SDL_GetTicks();                     // Current tick count
            long interval     = tickCurrent - tickPrevAvaraging;    // Interval

            // Calculate new average FPS
            fpsAverage = (interval == 0) ? 0.0f : (cast(float)countFPSAveragingFrame / cast(float)interval * 1000.0f);

            tickPrevAvaraging = tickCurrent;     // Next time's tickPrevious = this time's tickCurrent
            counterFPSAveragingFrame = 0;        // Reset counter
        }
    }
}






//--------------------------------------------------------------------------------
// Fiber Managing module
//--------------------------------------------------------------------------------
public class FiberModule
{
//  Values
//--------------------------------------------------
private:
    list!(IFiber)   listFiber;                // Fiber list
    int             indexScene = 0;           // Scene index
    int             indexNextScene = 0;       // Next scene index
    bool            flagSceneChanged = false; // Will scene change in next frame ?
    bool            flagHalted = false;       // Is halted ?


//  Constructor
//--------------------------------------------------
public:
    this()
    {
        listFiber = new list!(IFiber);
    }


//  Functions
//--------------------------------------------------
public:
    // add new fiber to the list
    void add(IFiber fiber)
    {
        listFiber.push_back(fiber);
    }


    // reset scene index
    void resetScene(int indexScene)
    {
        indexScene       = indexScene;
        indexNextScene   = indexScene;
        flagSceneChanged = false;
    }


    // change scene index in the head of next frame
    void changeScene(int indexNextScene)
    {
        indexNextScene   = indexNextScene;
        flagSceneChanged = true;
    }


    // system halt
    void halt()
    {
        flagHalted = true;
    }


    // is system halted ?
    bool isHalted()
    {
        return flagHalted;
    }


    // refer scene index
    int getScene()
    {
        return indexScene;
    }


//  Functions call from main routines
//--------------------------------------------------
private:
    // call all IFiber.enter()
    void enter_all()
    {
        foreach (IFiber fiber; listFiber) {
            fiber.enter();
        }
    }


    // call all IFiber.exit()
    void exit_all()
    {
        foreach_reverse (IFiber fiber; listFiber) {
            fiber.exit();
        }
    }


    // call all IFiber.enterScene()
    void enter_scene_all()
    {
        foreach (IFiber fiber; listFiber) {
            if (fiber.isActive()) {
                fiber.enterScene();
            }
        }
    }


    // call all IFiber.exitScene()
    void exit_scene_all()
    {
        foreach_reverse (IFiber fiber; listFiber) {
            if (fiber.isActive()) {
                fiber.exitScene();
            }
        }
    }


    // call all IFiber.run()
    void run_all()
    {
        // scene change
        if (flagSceneChanged) {
            exit_scene_all();
            indexScene = indexNextScene;
            enter_scene_all();
            flagSceneChanged = false;
        }
        
        // run all active fiber
        foreach (IFiber fiber; listFiber) {
            if (fiber.isActive()) {
                fiber.run();
            }
        }
    }


    // call all IFiber.draw()
    void draw_all()
    {
        foreach (IFiber fiber; listFiber) {
            if (fiber.isActive()) {
                fiber.draw();
            }
        }
    }
}






//--------------------------------------------------------------------------------
// Controler Managing module
//--------------------------------------------------------------------------------
public class ControlModule
{
//  Values
//--------------------------------------------------
private:
    list!(CControler)   listControler;      // controler list
    CControler          ctrlDefault;        // default controler

    SDLKey              sdlkeySystemHalt;   // system halt key
    ubyte*              tableStatusFlag;    // table of pressed key flags


//  Constructor
//--------------------------------------------------
public:
    this()
    {
        // controler list
        listControler = new list!(CControler);
        
        // add default controler to list
        ctrlDefault = add(new CControler);
        
        // default key assign
        ctrlDefault.clearAssign();
        ctrlDefault.assign(SDLK_LEFT,   PAD.LEFT);
        ctrlDefault.assign(SDLK_RIGHT,  PAD.RIGHT);
        ctrlDefault.assign(SDLK_UP,     PAD.UP);
        ctrlDefault.assign(SDLK_DOWN,   PAD.DOWN);
        ctrlDefault.assign(SDLK_a,      PAD.LEFT);
        ctrlDefault.assign(SDLK_d,      PAD.RIGHT);
        ctrlDefault.assign(SDLK_w,      PAD.UP);
        ctrlDefault.assign(SDLK_s,      PAD.DOWN);
        ctrlDefault.assign(SDLK_z,      PAD.BUTTON0);
        ctrlDefault.assign(SDLK_x,      PAD.BUTTON1);
        ctrlDefault.assign(SDLK_c,      PAD.BUTTON2);
        ctrlDefault.assign(SDLK_LCTRL,  PAD.BUTTON0);
        ctrlDefault.assign(SDLK_LSHIFT, PAD.BUTTON1);
        ctrlDefault.assign(SDLK_LALT,   PAD.BUTTON2);
        ctrlDefault.assign(SDLK_COMMA,  PAD.BUTTON0);
        ctrlDefault.assign(SDLK_PERIOD, PAD.BUTTON1);
        ctrlDefault.assign(SDLK_SLASH,  PAD.BUTTON2);
        
        // set default system halt key
        sdlkeySystemHalt = SDLK_F1;
    }


//  Functions
//--------------------------------------------------
//  setting
//--------------------------------------------------
public:
    // reset all contolers
    void reset()
    {
        foreach (CControler ctrl; listControler) {
            ctrl.reset();
        }
    }


    // update joystick list
    void updateJoystickList()
    {
        foreach (CControler ctrl; listControler) {
            ctrl.closeJoystick();
            ctrl.openJoystick();
        }
    }


    // add new contoler
    CControler add(CControler ctrl)
    {
        ctrl._internal_set_id(listControler.size());
        listControler.push_back(ctrl);
        return ctrl;
    }

    
    // Set system halt key
    void setSystemHalt(SDLKey key)
    {
        sdlkeySystemHalt = key;
    }


    // Get default controler instance
    CControler getDefaultControler()
    {
        return ctrlDefault;
    }


//  reference
//--------------------------------------------------
public:
    // Get table of SDLKey flags
    ubyte* getSDLKeyFlags()
    {
        return (moduleControl.tableStatusFlag);
    }


    // Check keyborad state
    bool isPressed(SDLKey key)
    {
        return (moduleControl.tableStatusFlag[key] == SDL_PRESSED);
    }


    // Check keyborad/joystick state
    bool isPressed(PAD key)
    {
        return (moduleControl.ctrlDefault.isPressed(key));
    }


    // Check keyborad/joystick state
    bool isHitted(PAD key)
    {
        return (moduleControl.ctrlDefault.isHitted(key));
    }


    // Get keyborad/joystick state frame counter
    int getFrameCounter(PAD key)
    {
        return (moduleControl.ctrlDefault.getFrameCounter(key));
    }


    // Get x and y input
    void getXY(inout int x, inout int y)
    {
        return (moduleControl.ctrlDefault.getXY(x, y));
    }

    
    

//  Functions call from main routines
//--------------------------------------------------
private:
    // is system halt key pressed ?
    bool is_system_halted()
    {
        return (tableStatusFlag != null && tableStatusFlag[sdlkeySystemHalt] == SDL_PRESSED);
    }


    // update table of SDLKey flag
    void update_keytable()
    {
        tableStatusFlag = SDL_GetKeyState(null);
        SDL_JoystickUpdate();
        foreach (CControler ctrl; listControler) {
            ctrl._internal_update_flag(tableStatusFlag);
        }
    }


    // update frame counter
    void update_counter()
    {
        foreach (CControler ctrl; listControler) {
            ctrl._internal_update_counter();
        }
    }
}


